/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.jcip.annotations.Immutable;

import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * This class provides static methods for parsing a context-free grammar written
 * in JSON syntax.
 *
 * <p>
 * For example:
 *
 * <pre>
 * {
 *     "startSymbol": "S",
 *     "rules": {
 *         "S": [ [ { "T?": "s" }, { "N": "A" }, { "N": "B" } ], [ ] ],
 *         "A": [ [ { "T": "a" }, { "N?": "A" } ], [ { "T": "f" } ] ],
 *         "B": [ [ { "T": "b" }, { "R?": "c-e" } ] ]
 *     }
 * }
 * </pre>
 *
 * For comparison, here is the same context-free grammar written in EBNF syntax:
 *
 * <pre>
 * S = [ "s" ], A, B |
 * A = "a", [ A ] | "f"
 * B = "b", [ ( "c" | "d" | "e" ) ]
 * </pre>
 *
 * @author Clemens Akens
 */
@Immutable
public final class JsonParser {

    /**
     * This enum represents a quantifier.
     */
    private enum Quantifier {

        /** This field represents the none quantifier. */
        None,

        /** This field represents the optional quantifier. */
        Optional
    }

    /**
     * This enum represents a symbol type.
     */
    private enum SymbolType {

        /** This field represents the non-terminal symbol type. */
        NonTerminal,

        /** This field represents the range symbol type. */
        Range,

        /** This field represents the terminal symbol type. */
        Terminal
    }

    /**
     * This static field stores the object mapper to generate the JSON object
     * representation from the JSON string representation of a context-free
     * grammar.
     */
    private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

    /**
     * This static field stores the regular expression to match a unicode
     * character range within the BMP.
     */
    private static final Pattern RANGE_PATTERN = Pattern.compile("(.)-(.)");

    /**
     * This static field stores the regular expression to match a symbol
     * descriptor.
     */
    private static final Pattern SYMBOL_DESCRIPTOR_PATTERN = Pattern.compile("(N|R|T)(\\?)?");

    static {
        JsonParser.OBJECT_MAPPER.configure(com.fasterxml.jackson.core.JsonParser.Feature.ALLOW_COMMENTS, true);
    }

    /**
     * This static method returns the context-free grammar parsed from the
     * specified JSON source.
     *
     * @param source
     *            The JSON source.
     * @return The context-free grammar parsed from the specified JSON source.
     * @throws ParseException
     *             Signals that a parsing error occured.
     */
    public static Grammar parse(final File source) throws ParseException {
        try {
            @SuppressWarnings("unchecked")
            final Map<String, Object> grammarObject = JsonParser.OBJECT_MAPPER.readValue(source, Map.class);

            return Grammar.create(JsonParser.parseStartSymbol(grammarObject), JsonParser.parseRules(grammarObject));
        } catch (final IOException exception) {
            throw new ParseException(exception.getMessage(), exception);
        }
    }

    /**
     * This static method returns the context-free grammar parsed from the
     * specified JSON source.
     *
     * @param source
     *            The JSON source.
     * @return The context-free grammar parsed from the specified JSON source.
     * @throws ParseException
     *             Signals that a parsing error occured.
     */
    public static Grammar parse(final InputStream source) throws ParseException {
        try {
            @SuppressWarnings("unchecked")
            final Map<String, Object> grammarObject = JsonParser.OBJECT_MAPPER.readValue(source, Map.class);

            return Grammar.create(JsonParser.parseStartSymbol(grammarObject), JsonParser.parseRules(grammarObject));
        } catch (final IOException exception) {
            throw new ParseException(exception.getMessage(), exception);
        }
    }

    /**
     * This static method returns the context-free grammar parsed from the
     * specified JSON source.
     *
     * @param source
     *            The JSON source.
     * @return The context-free grammar parsed from the specified JSON source.
     * @throws ParseException
     *             Signals that a parsing error occured.
     */
    public static Grammar parse(final Reader source) throws ParseException {
        try {
            @SuppressWarnings("unchecked")
            final Map<String, Object> grammarObject = JsonParser.OBJECT_MAPPER.readValue(source, Map.class);

            return Grammar.create(JsonParser.parseStartSymbol(grammarObject), JsonParser.parseRules(grammarObject));
        } catch (final IOException exception) {
            throw new ParseException(exception.getMessage(), exception);
        }
    }

    /**
     * This static method returns the context-free grammar parsed from the
     * specified JSON content.
     *
     * @param content
     *            The JSON content.
     * @return The context-free grammar parsed from the specified JSON content.
     * @throws ParseException
     *             Signals that a parsing error occured.
     */
    public static Grammar parse(final String content) throws ParseException {
        try {
            @SuppressWarnings("unchecked")
            final Map<String, Object> grammarObject = JsonParser.OBJECT_MAPPER.readValue(content, Map.class);

            return Grammar.create(JsonParser.parseStartSymbol(grammarObject), JsonParser.parseRules(grammarObject));
        } catch (final IOException exception) {
            throw new ParseException(exception.getMessage(), exception);
        }
    }

    /**
     * This static method returns the context-free grammar parsed from the
     * specified JSON source.
     *
     * @param source
     *            The JSON source.
     * @return The context-free grammar parsed from the specified JSON source.
     * @throws ParseException
     *             Signals that a parsing error occured.
     */
    public static Grammar parse(final URL source) throws ParseException {
        try {
            @SuppressWarnings("unchecked")
            final Map<String, Object> grammarObject = JsonParser.OBJECT_MAPPER.readValue(source, Map.class);

            return Grammar.create(JsonParser.parseStartSymbol(grammarObject), JsonParser.parseRules(grammarObject));
        } catch (final IOException exception) {
            throw new ParseException(exception.getMessage(), exception);
        }
    }

    /**
     * This static method returns the rules object from the specified grammar
     * object.
     *
     * @param grammarObject
     *            The grammar object.
     * @return The rules object from the specified grammar object.
     * @throws ParseException
     *             Signals that the grammar object is malformed.
     */
    @SuppressWarnings("unchecked")
    private static Map<String, List<List<Map<String, String>>>> getRulesObject(final Map<String, Object> grammarObject) throws ParseException {
        if (grammarObject.get("rules") == null) {
            throw new ParseException("Undefined rules");
        }

        Map<String, List<List<Map<String, String>>>> rulesObject = null;

        try {
            rulesObject = (Map<String, List<List<Map<String, String>>>>) grammarObject.get("rules");
        } catch (final ClassCastException exception) {
            throw new ParseException("Unrecognized rules: " + grammarObject.get("rules"));
        }

        for (final Entry<String, List<List<Map<String, String>>>> entry : rulesObject.entrySet()) {
            if (entry.getKey().isEmpty()) {
                throw new ParseException("Unrecognized non-terminal: ");
            }

            List<List<Map<String, String>>> rulesObjectForKey = null;

            try {
                rulesObjectForKey = entry.getValue();
            } catch (final ClassCastException exception) {
                throw new ParseException("Unrecognized rules for non-terminal " + entry.getKey() + ": " + entry.getValue());
            }

            if (rulesObjectForKey.isEmpty()) {
                throw new ParseException("Undefined rules for non-terminal " + entry.getKey());
            }

            for (int i = 0; i < rulesObjectForKey.size(); ++i) {
                List<Map<String, String>> ruleObject = null;

                try {
                    ruleObject = rulesObjectForKey.get(i);
                } catch (final ClassCastException exception) {
                    throw new ParseException("Unrecognized rule for non-terminal " + entry.getKey() + ": " + rulesObjectForKey.get(i));
                }

                for (int j = 0; j < ruleObject.size(); ++j) {
                    Map<String, String> symbolObject = null;

                    try {
                        symbolObject = ruleObject.get(j);
                    } catch (final ClassCastException exception) {
                    }

                    if ((symbolObject == null) || symbolObject.isEmpty() || (symbolObject.size() > 1)) {
                        throw new ParseException("Unrecognized symbol: " + ruleObject.get(j));
                    }

                    String name = null;

                    try {
                        name = symbolObject.values().iterator().next();
                    } catch (final ClassCastException exception) {
                    }

                    if ((name == null) || name.isEmpty()) {
                        throw new ParseException("Unrecognized symbol name: " + (Object) symbolObject.values().iterator().next());
                    }
                }
            }
        }

        return rulesObject;
    }

    /**
     * This static method returns the quantifier parsed from the specified
     * symbol descriptor.
     *
     * @param symbolDescriptor
     *            The symbol descriptor.
     * @return The quantifier parsed from the specified symbol descriptor.
     */
    private static Quantifier parseQuantifier(final String symbolDescriptor) {
        final Matcher matcher = JsonParser.SYMBOL_DESCRIPTOR_PATTERN.matcher(symbolDescriptor);

        Quantifier quantifier = Quantifier.None;

        if (matcher.matches() && (matcher.group(2) != null)) {
            quantifier = Quantifier.Optional;
        }

        return quantifier;
    }

    /**
     * This static method returns the set of all production rules parsed from
     * the specified grammar object.
     *
     * @param grammarObject
     *            The grammar object.
     * @return The set of all production rules parsed from the specified grammar
     *         object.
     * @throws ParseException
     *             Signals that the grammar object is malformed.
     */
    private static Set<Rule> parseRules(final Map<String, Object> grammarObject) throws ParseException {
        final Set<Rule> rules = new HashSet<Rule>();
        final Map<String, List<List<Map<String, String>>>> rulesObject = JsonParser.getRulesObject(grammarObject);

        final Set<NonTerminal> nonTerminals = new HashSet<NonTerminal>();

        for (final String key : rulesObject.keySet()) {
            nonTerminals.add(NonTerminal.create(key));
        }

        for (final Entry<String, List<List<Map<String, String>>>> entry : rulesObject.entrySet()) {
            for (final List<Map<String, String>> ruleObject : entry.getValue()) {
                final Set<List<Symbol>> rightHandSides = new HashSet<List<Symbol>>();

                rightHandSides.add(new ArrayList<Symbol>());

                for (final Map<String, String> symbolObject : ruleObject) {
                    final String symbolDescriptor = symbolObject.keySet().iterator().next();
                    final String name = symbolObject.values().iterator().next();
                    final Quantifier quantifier = JsonParser.parseQuantifier(symbolDescriptor);
                    final SymbolType symbolType = JsonParser.parseSymbolType(symbolDescriptor);

                    final Set<List<Symbol>> newRightHandSides = new HashSet<List<Symbol>>();

                    for (List<Symbol> rightHandSide : rightHandSides) {
                        if (Quantifier.Optional.equals(quantifier)) {
                            rightHandSide = new ArrayList<Symbol>(rightHandSide);

                            newRightHandSides.add(rightHandSide);
                        }

                        if (SymbolType.NonTerminal.equals(symbolType)) {
                            rightHandSide.add(NonTerminal.create(name));
                        } else if (SymbolType.Terminal.equals(symbolType)) {
                            rightHandSide.add(Terminal.create(name));
                        } else {
                            final Matcher matcher = JsonParser.RANGE_PATTERN.matcher(name);

                            if (matcher.matches() && (matcher.group(1).codePointAt(0) < matcher.group(2).codePointAt(0))) {
                                final NonTerminal newNonTerminal = NonTerminal.createUnique(name, nonTerminals);

                                final char start = matcher.group(1).charAt(0);
                                final char end = matcher.group(2).charAt(0);

                                for (char i = start; i <= end; ++i) {
                                    rules.add(Rule.create(newNonTerminal, Terminal.create(Character.toString(i))));
                                }

                                rightHandSide.add(newNonTerminal);
                            } else {
                                throw new ParseException("Unrecognized range: " + name);
                            }
                        }
                    }

                    rightHandSides.addAll(newRightHandSides);
                }

                for (final List<Symbol> rightHandSide : rightHandSides) {
                    rules.add(Rule.create(NonTerminal.create(entry.getKey()), rightHandSide));
                }
            }
        }

        return rules;
    }

    /**
     * This static method returns the start symbol parsed from the specified
     * grammar object.
     *
     * @param grammarObject
     *            The grammar object.
     * @return The start symbol parsed from the specified grammar object.
     * @throws ParseException
     *             Signals that the grammar object is malformed.
     */
    private static NonTerminal parseStartSymbol(final Map<String, Object> grammarObject) throws ParseException {
        if (grammarObject.get("startSymbol") == null) {
            throw new ParseException("Undefined start symbol");
        }

        String name = null;

        try {
            name = (String) grammarObject.get("startSymbol");
        } catch (final ClassCastException exception) {
        }

        if ((name == null) || name.isEmpty()) {
            throw new ParseException("Unrecognized start symbol: " + grammarObject.get("startSymbol"));
        }

        return NonTerminal.create(name);
    }

    /**
     * This static method returns the symbol type parsed from the specified
     * symbol descriptor.
     *
     * @param symbolDescriptor
     *            The symbol descriptor.
     * @return The symbol type parsed from the specified symbol descriptor.
     * @throws ParseException
     *             Signals that the symbol descriptor is unrecognized.
     */
    private static SymbolType parseSymbolType(final String symbolDescriptor) throws ParseException {
        final Matcher matcher = JsonParser.SYMBOL_DESCRIPTOR_PATTERN.matcher(symbolDescriptor);

        SymbolType symbolType = null;

        if (matcher.matches()) {
            final String result = matcher.group(1);

            if (result.equals("N")) {
                symbolType = SymbolType.NonTerminal;
            } else if (result.equals("R")) {
                symbolType = SymbolType.Range;
            } else {
                symbolType = SymbolType.Terminal;
            }
        } else {
            throw new ParseException("Unrecognized symbol descriptor: " + symbolDescriptor);
        }

        return symbolType;
    }

    /**
     * This constructor instantiates a new JSON parser.
     */
    private JsonParser() {
    }
}
