/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.Set;

import net.jcip.annotations.Immutable;

/**
 * This class represents a non-terminal symbol in a context-free grammar.
 *
 * <p>
 * Non-terminal symbols, or just non-terminals, are the symbols which can be
 * replaced; thus there are strings composed of some combination of terminal and
 * non-terminal symbols. They may also be called simply syntactic variables.
 *
 * <p>
 * A context-free grammar includes a start symbol, a designated member of the
 * set of non-terminals from which all the strings in the language may be
 * derived by successive applications of the production rules. In fact, the
 * language defined by a grammar is precisely the set of terminal strings that
 * can be so derived.
 *
 * <p>
 * <i>For more information see the Wikipedia article: <a
 * href="http://en.wikipedia.org/wiki/Nonterminal">Terminal and nonterminal
 * symbols</a></i>
 *
 * @author Clemens Akens
 */
@Immutable
public final class NonTerminal extends Symbol {

    /**
     * This static method creates a new non-terminal symbol.
     *
     * @param name
     *            The non-empty string representation of the new non-terminal
     *            symbol.
     * @return The new non-terminal symbol.
     */
    public static NonTerminal create(final String name) {
        return new NonTerminal(name);
    }

    /**
     * This static method creates a new unique non-terminal symbol.
     *
     * @param name
     *            The non-empty string representation of the new non-terminal
     *            symbol.
     * @param existingNonTerminals
     *            The set of existing non-terminal symbols.
     * @return The new unique non-terminal symbol.
     */
    public static NonTerminal createUnique(final String name, final Set<NonTerminal> existingNonTerminals) {
        int index = 0;
        NonTerminal nonTerminal = NonTerminal.create(name);

        while (existingNonTerminals.contains(nonTerminal)) {
            nonTerminal = NonTerminal.create(name + index++);
        }

        return nonTerminal;
    }

    /**
     * This constructor instantiates a new non-terminal symbol.
     *
     * @param name
     *            The non-empty string representation of the new non-terminal
     *            symbol.
     */
    private NonTerminal(final String name) {
        super(name);
    }
}
