/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

/**
 * This enum represents some properties of a context-free grammar.
 *
 * @author Clemens Akens
 */
public enum Property {

    /**
     * This property indicates that the context-free grammar does not contain
     * non-productive production rules or unreachable non-terminal symbols.
     */
    Clean("non-productive production rules or unreachable non-terminal symbols"),

    /**
     * This property indicates that the context-free grammar does not contain
     * terminal symbols with the same string representation as non-terminal
     * symbols.
     */
    Disjoint("terminal symbols with the same string representation as non-terminal symbols"),

    /**
     * This property indicates that the context-free grammar does not contain
     * nullable non-terminal symbols. Only the start symbol can be nullable.
     */
    EmptyFree("nullable non-terminal symbols"),

    /**
     * This property indicates that the context-free grammar does not contain
     * left-recursive non-terminal symbols.
     */
    LeftRecursionFree("left-recursive non-terminal symbols"),

    /**
     * This property indicates that the context-free grammar does not contain
     * unit production rules.
     */
    UnitFree("unit production rules");

    /** This field stores the description of this property. */
    private final String description;

    /**
     * This constructor instantiates a new property.
     *
     * @param description
     *            The description of this property.
     */
    private Property(final String description) {
        this.description = description;
    }

    /**
     * This method returns the description of this property.
     *
     * @return The description of this property.
     */
    String getDescription() {
        return this.description;
    }
}
