/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import net.jcip.annotations.Immutable;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;

/**
 * This class represents a production rule in a context-free grammar.
 *
 * <p>
 * Production rules, or just rules, are rewrite rules specifying a symbol
 * substitution that can be recursively performed to generate new symbol
 * sequences.
 *
 * <p>
 * The left-hand side of each production rule in a context-free grammar consists
 * of only a single non-terminal symbol.
 *
 * <p>
 * <i>For more information see the Wikipedia article: <a
 * href="http://en.wikipedia.org/wiki/Production_(computer_science)">Production
 * (computer science)</a></i>
 *
 * @author Clemens Akens
 */
@Immutable
public final class Rule {

    /**
     * This static field stores the object mapper to generate the JSON string
     * representation from the JSON object representation of a production rule.
     */
    private static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();

    /**
     * This static method creates a new production rule.
     *
     * @param leftHandSide
     *            The left-hand side of the new production rule.
     * @param rightHandSide
     *            The (possibly empty) right-hand side of the new production
     *            rule.
     * @return The new production rule.
     */
    public static Rule create(final NonTerminal leftHandSide, final Collection<? extends Symbol> rightHandSide) {
        return new Rule(leftHandSide, ImmutableList.copyOf(rightHandSide));
    }

    /**
     * This static method creates a new production rule.
     *
     * @param leftHandSide
     *            The left-hand side of the new production rule.
     * @param rightHandSide
     *            The (possibly empty) right-hand side of the new production
     *            rule.
     * @return The new production rule.
     */
    public static Rule create(final NonTerminal leftHandSide, final Symbol... rightHandSide) {
        return new Rule(leftHandSide, ImmutableList.copyOf(rightHandSide));
    }

    /** This field stores the left-hand side of this production rule. */
    private final NonTerminal leftHandSide;

    /** This field stores the right-hand side of this production rule. */
    private final ImmutableList<? extends Symbol> rightHandSide;

    /**
     * This constructor instantiates a new production rule.
     *
     * @param leftHandSide
     *            The left-hand side of the new production rule.
     * @param rightHandSide
     *            The (possibly empty) right-hand side of the new production
     *            rule.
     */
    private Rule(final NonTerminal leftHandSide, final ImmutableList<? extends Symbol> rightHandSide) {
        this.leftHandSide = Preconditions.checkNotNull(leftHandSide, "leftHandSide is null");
        this.rightHandSide = rightHandSide;
    }

    /**
     * This method indicates whether some other object is "equal to" this one.
     *
     * @param object
     *            The other object.
     * @return True, if this object is equal to the other object.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(final Object object) {
        if (object == this) {
            return true;
        }

        if (object == null) {
            return false;
        }

        if (object.getClass() != this.getClass()) {
            return false;
        }

        final Rule other = (Rule) object;

        return Objects.equal(other.leftHandSide, this.leftHandSide) && Objects.equal(other.rightHandSide, this.rightHandSide);
    }

    /**
     * This method returns the left-hand side of this production rule.
     *
     * @return The left-hand side of this production rule.
     */
    public NonTerminal getLeftHandSide() {
        return this.leftHandSide;
    }

    /**
     * This method returns the right-hand side of this production rule.
     *
     * @return The right-hand side of this production rule.
     */
    public ImmutableList<? extends Symbol> getRightHandSide() {
        return this.rightHandSide;
    }

    /**
     * This method returns the hash code value for this object.
     *
     * @return The hash code value for this object.
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode() {
        return Objects.hashCode(this.leftHandSide, this.rightHandSide);
    }

    /**
     * This method returns the JSON object representation of this production
     * rule.
     *
     * @return The JSON object representation of this production rule.
     */
    public Object toObject() {
        final List<Object> object = new ArrayList<Object>();

        if (this.rightHandSide != null) { // EqualsVerifier: Non-nullity
            for (final Symbol symbol : this.rightHandSide) {
                object.add(symbol.toObject());
            }
        }

        return object;
    }

    /**
     * This method returns the JSON string representation of this production
     * rule.
     *
     * @return The JSON string representation of this production rule.
     */
    @Override
    public String toString() {
        try {
            return Rule.OBJECT_MAPPER.writeValueAsString(this.toObject());
        } catch (final JsonProcessingException exception) {
            return "";
        }
    }
}
