/*
 * Copyright 2012 Clemens Akens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.clebert.jlingua;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import net.jcip.annotations.Immutable;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.Multimap;

/**
 * This class provides static methods for transforming a context-free grammar.
 *
 * @author Clemens Akens
 */
@Immutable
public final class Transformer {

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all non-productive production rules and unreachable non-terminal
     * symbols are removed.
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all
     *         non-productive production rules and unreachable non-terminal
     *         symbols are removed.
     */
    public static Grammar cleanUp(final Grammar grammar) {
        return Transformer.removeUnreachableNonTerminals(Transformer.removeNonProductiveRules(grammar));
    }

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all empty production rules and thus all loops concealed by empty
     * production rules are eliminated.
     *
     * <p>
     * A loop may only become visible when certain non-terminal symbols produce
     * empty.<br>
     * The resulting grammar is cleaned up and can contain unit production
     * rules.
     *
     * <p>
     * Introduces a new nullable start symbol and in addition a production rule
     * with a right-hand side consisting of the old start symbol, if the old
     * start symbol is nullable and it occurs on the right-hand side of a
     * productive production rule with a reachable non-terminal symbol as its
     * left-hand-side.
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all empty
     *         production rules, and thus all loops concealed by empty
     *         production rules are eliminated.
     */
    public static Grammar eliminateEmptyRules(final Grammar grammar) {
        final Grammar cleanGrammar = Transformer.cleanUp(grammar);
        final ImmutableSet<NonTerminal> nullableNonTerminals = Utils.computeNullableNonTerminals(cleanGrammar);

        if (nullableNonTerminals.isEmpty()) {
            return cleanGrammar;
        }

        final Set<Rule> rules = new HashSet<Rule>(cleanGrammar.getRules());

        for (final NonTerminal nonTerminal : nullableNonTerminals) {
            final Set<Rule> unprocessedRules = new HashSet<Rule>(rules);

            while (!unprocessedRules.isEmpty()) {
                final Set<Rule> newRules = new HashSet<Rule>();

                for (final Iterator<Rule> iterator = unprocessedRules.iterator(); iterator.hasNext();) {
                    final Rule rule = iterator.next();

                    iterator.remove();

                    for (int index = 0; index < rule.getRightHandSide().size(); ++index) {
                        final Symbol symbol = rule.getRightHandSide().get(index);

                        if (symbol.equals(nonTerminal)) {
                            final List<Symbol> rightHandSide = new ArrayList<Symbol>(rule.getRightHandSide());

                            rightHandSide.remove(index);

                            final Rule newRule = Rule.create(rule.getLeftHandSide(), rightHandSide);

                            if (!rules.contains(newRule)) {
                                newRules.add(newRule);
                            }
                        }
                    }
                }

                rules.addAll(newRules);
                unprocessedRules.addAll(newRules);
            }
        }

        for (final Iterator<Rule> iterator = rules.iterator(); iterator.hasNext();) {
            if (iterator.next().getRightHandSide().isEmpty()) {
                iterator.remove();
            }
        }

        NonTerminal startSymbol = cleanGrammar.getStartSymbol();

        if (nullableNonTerminals.contains(startSymbol)) {
            if (Utils.computeDirectReachabilityRelation(cleanGrammar).containsValue(startSymbol)) {
                startSymbol = NonTerminal.createUnique(startSymbol.getName(), cleanGrammar.getNonTerminals());

                rules.add(Rule.create(startSymbol, cleanGrammar.getStartSymbol()));
            }

            rules.add(Rule.create(startSymbol));
        }

        return Transformer.cleanUp(Grammar.create(startSymbol, rules));
    }

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all empty production rules, then all unit production rules and
     * then all left-recursive non-terminal symbols are eliminated.
     *
     * <p>
     * The resulting grammar is cleaned up and can contain empty production
     * rules and unit production rules.
     *
     * <h4>Rule 1</h4>
     *
     * If a terminal symbol or non-left-recursive non-terminal symbol X is a
     * left corner of a retained left-recursive non-terminal symbol A in the
     * original grammar, add A → X A-X to the transformed grammar.
     *
     * <h4>Rule 2</h4>
     *
     * If B is a left-recursive left corner of a retained left-recursive
     * non-terminal symbol A and B → X β is a production rule of the original
     * grammar, add A-X → β A-B to the transformed grammar.
     *
     * <h4>Rule 3</h4>
     *
     * If X is a left corner of a retained left-recursive non-terminal symbol A
     * and A → X β is a production rule of the original grammar, add A-X → β to
     * the transformed grammar.
     *
     * <h4>Rule 4</h4>
     *
     * If A is a non-left-recursive non-terminal symbol and A → β is a
     * production rule of the original grammar, add A → β to the transformed
     * grammar.
     *
     * <p>
     * <i>For more information on the algorithm, see Moore, Robert C. (May
     * 2000). "Removing Left Recursion from Context-Free Grammars". 6th Applied
     * Natural Language Processing Conference: 249–255.</i>
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all empty
     *         production rules, then all unit production rules and then all
     *         left-recursive non-terminal symbols are eliminated.
     */
    public static Grammar eliminateLeftRecursion(final Grammar grammar) {
        final Grammar unitFreeGrammar = Transformer.eliminateUnitRules(grammar);
        final ImmutableSetMultimap<NonTerminal, NonTerminal> leftCornerRelation = Utils.computeLeftCornerRelation(unitFreeGrammar);
        final ImmutableSet<NonTerminal> leftRecursiveNonTerminals = Utils.computeLeftRecursiveNonTerminals(unitFreeGrammar, leftCornerRelation);
        final ImmutableSet<NonTerminal> retainedNonTerminals = Utils.computeRetainedNonTerminals(unitFreeGrammar);
        final Set<Rule> rules = new HashSet<Rule>();

        for (final NonTerminal nonTerminal : unitFreeGrammar.getNonTerminals()) {
            if (!leftRecursiveNonTerminals.contains(nonTerminal)) {
                rules.addAll(unitFreeGrammar.getRulesByNonTerminal().get(nonTerminal));
            } else if (retainedNonTerminals.contains(nonTerminal)) {
                final Set<Rule> leftCornerRules = new HashSet<Rule>();

                for (final Rule rule : unitFreeGrammar.getRulesByNonTerminal().get(nonTerminal)) {
                    final ImmutableList<? extends Symbol> rightHandSide = rule.getRightHandSide();
                    final NonTerminal newNonTerminal = Transformer.createLinkedNonTerminal(unitFreeGrammar, nonTerminal, rightHandSide.get(0));

                    rules.add(Rule.create(newNonTerminal, rightHandSide.subList(1, rightHandSide.size())));
                }

                for (final NonTerminal leftCornerNonTerminal : leftCornerRelation.get(nonTerminal)) {
                    leftCornerRules.addAll(unitFreeGrammar.getRulesByNonTerminal().get(leftCornerNonTerminal));
                }

                for (final Rule rule1 : leftCornerRules) {
                    final Symbol leftCornerSymbol = rule1.getRightHandSide().get(0);
                    final NonTerminal newNonTerminal1 = Transformer.createLinkedNonTerminal(unitFreeGrammar, nonTerminal, leftCornerSymbol);

                    if (leftRecursiveNonTerminals.contains(leftCornerSymbol)) {
                        for (final Rule rule2 : unitFreeGrammar.getRulesByNonTerminal().get((NonTerminal) leftCornerSymbol)) {
                            final ImmutableList<? extends Symbol> rightHandSide = rule2.getRightHandSide();
                            final NonTerminal newNonTerminal2 = Transformer.createLinkedNonTerminal(unitFreeGrammar, nonTerminal, rightHandSide.get(0));
                            final List<Symbol> newRightHandSide = new ArrayList<Symbol>();

                            newRightHandSide.addAll(rightHandSide.subList(1, rightHandSide.size()));
                            newRightHandSide.add(newNonTerminal1);

                            rules.add(Rule.create(newNonTerminal2, newRightHandSide));
                        }
                    } else {
                        rules.add(Rule.create(nonTerminal, leftCornerSymbol, newNonTerminal1));
                    }
                }
            }
        }

        return Transformer.cleanUp(Grammar.create(unitFreeGrammar.getStartSymbol(), rules));
    }

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all empty production rules and then all unit production rules
     * are eliminated.
     *
     * <p>
     * The algorithm can produce unreachable non-terminal symbols, therefore the
     * resulting grammar is cleaned up and is not infinitely ambiguous any more.
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all empty
     *         production rules and then all unit production rules are
     *         eliminated.
     */
    public static Grammar eliminateUnitRules(final Grammar grammar) {
        final Grammar emptyFreeGrammar = Transformer.eliminateEmptyRules(grammar);
        final Multimap<NonTerminal, Rule> rulesByNonTerminal = HashMultimap.create(emptyFreeGrammar.getRulesByNonTerminal());

        for (final NonTerminal nonTerminal : emptyFreeGrammar.getNonTerminals()) {
            final Set<Rule> newRules = new HashSet<Rule>();

            for (final Iterator<Rule> iterator = rulesByNonTerminal.values().iterator(); iterator.hasNext();) {
                final Rule rule1 = iterator.next();

                if (rule1.getRightHandSide().size() == 1) {
                    final Symbol symbol = rule1.getRightHandSide().get(0);

                    if (symbol.equals(nonTerminal)) {
                        iterator.remove();

                        if (!rule1.getLeftHandSide().equals(nonTerminal)) {
                            for (final Rule rule2 : rulesByNonTerminal.get(nonTerminal)) {
                                if ((rule2.getRightHandSide().size() != 1) || !rule2.getRightHandSide().get(0).equals(nonTerminal)) {
                                    newRules.add(Rule.create(rule1.getLeftHandSide(), rule2.getRightHandSide()));
                                }
                            }
                        }
                    }
                }
            }

            for (final Rule newRule : newRules) {
                rulesByNonTerminal.put(newRule.getLeftHandSide(), newRule);
            }
        }

        return Transformer.cleanUp(Grammar.create(emptyFreeGrammar.getStartSymbol(), rulesByNonTerminal.values()));
    }

    /**
     * This static method creates a new linked non-terminal symbol.
     *
     * @param grammar
     *            The context-free grammar.
     * @param nonTerminal
     *            The non-terminal symbol to be linked.
     * @param symbol
     *            The symbol to be linked against.
     * @return The new linked non-terminal symbol.
     */
    private static NonTerminal createLinkedNonTerminal(final Grammar grammar, final NonTerminal nonTerminal, final Symbol symbol) {
        final String name;

        if (symbol.isNonTerminal()) {
            name = nonTerminal.getName() + "-{" + symbol.getName() + "}";
        } else {
            name = nonTerminal.getName() + "-[" + symbol.getName() + "]";
        }

        return NonTerminal.createUnique(name, grammar.getNonTerminals());
    }

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all non-productive production rules are removed.
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all
     *         non-productive production rules are removed.
     */
    private static Grammar removeNonProductiveRules(final Grammar grammar) {
        return Grammar.create(grammar.getStartSymbol(), Utils.computeProductiveRulesByNonTerminal(grammar).values());
    }

    /**
     * This static method returns a copy of the specified context-free grammar
     * in which all unreachable non-terminal symbols are removed.
     *
     * <p>
     * First removing unreachable non-terminal symbols and then removing
     * non-productive production rules may produce a grammar which again
     * contains unreachable non-terminal symbols.
     *
     * @param grammar
     *            The context-free grammar.
     * @return A copy of the specified context-free grammar in which all
     *         unreachable non-terminal symbols are removed.
     */
    private static Grammar removeUnreachableNonTerminals(final Grammar grammar) {
        final Set<Rule> rules = new HashSet<Rule>();

        for (final NonTerminal nonTerminal : Utils.computeReachableNonTerminals(grammar)) {
            rules.addAll(grammar.getRulesByNonTerminal().get(nonTerminal));
        }

        return Grammar.create(grammar.getStartSymbol(), rules);
    }

    /**
     * This constructor instantiates a new transformer.
     */
    private Transformer() {
    }
}
